@@ -23,6 +23,32 @@ def get_EF(input_size, dim, bias=True):
     torch.nn.init.xavier_normal_(lin.weight)
     return lin
 
+class Residual(nn.Module):
+    """
+    Implemenation taken from
+    https://github.com/lucidrains/sinkhorn-transformer/blob/master/sinkhorn_transformer/sinkhorn_transformer.py
+    """
+    def __init__(self, fn):
+        super(Residual, self).__init__()
+        self.fn = fn
+
+    def forward(self, tensor, **kwargs):
+        return tensor + self.fn(tensor, **kwargs)
+
+class PreNorm(nn.Module):
+    """
+    Implemenation taken from
+    https://github.com/lucidrains/sinkhorn-transformer/blob/master/sinkhorn_transformer/sinkhorn_transformer.py
+    """
+    def __init__(self, channels, fn):
+        super(PreNorm, self).__init__()
+        self.fn = fn
+        self.norm = nn.LayerNorm(channels)
+
+    def forward(self, tensor, **kwargs):
+        tensor = self.norm(tensor)
+        return self.fn(tensor, **kwargs)
+
 class PositionalEmbedding(nn.Module):
     """
     Standard positional embedding.
@@ -93,7 +119,7 @@ class LinearAttentionHead(nn.Module):
         Assume Q, K, V have same dtype
         E, F are `nn.Linear` modules
         """
-        K = torch.transpose(K, 1, 2)
+        K = K.transpose(1,2)
         if not self.full_attention:
             K = self.E(K)
         Q = torch.matmul(Q, K)
@@ -108,9 +134,9 @@ class LinearAttentionHead(nn.Module):
         P_bar = self.dropout(P_bar)
 
         if not self.full_attention:
-            V = torch.transpose(V, 1, 2)
+            V = V.transpose(1,2)
             V = self.F(V)
-            V = torch.transpose(V, 1, 2)
+            V = V.transpose(1,2)
         out_tensor = torch.matmul(P_bar, V)
 
         return out_tensor
@@ -120,22 +146,24 @@ class MHAttention(nn.Module):
     Multihead attention, with each head being a Linformer Head
     This feeds directly into a feed forward head
     """
-    def __init__(self, input_size, dim, channels, dim_k, nhead, dropout, activation, checkpoint_level, parameter_sharing, E_proj, F_proj, full_attention, w_o_intermediate_dim=None):
+    def __init__(self, input_size, dim, channels, dim_k, nhead, dropout, activation, checkpoint_level, parameter_sharing, E_proj, F_proj, full_attention, w_o_intermediate_dim=None, decoder_mode=False):
         super(MHAttention, self).__init__()
         self.heads = nn.ModuleList()
         self.input_size = input_size
         self.dim_k = dim_k
+        self.channels = channels
         self.checkpoint_level = checkpoint_level
         self.w_o_intermediate_dim = w_o_intermediate_dim
         if parameter_sharing != "layerwise":
             E_proj = get_EF(input_size, dim_k)
             F_proj = get_EF(input_size, dim_k) if parameter_sharing == "none" or parameter_sharing == "headwise" else E_proj
 
+        self.decoder_mode = decoder_mode
         self.to_q = nn.ModuleList()
         self.to_k = nn.ModuleList()
         self.to_v = nn.ModuleList()
 
-        for head in range(nhead):
+        for _ in range(nhead):
             if parameter_sharing == "none":
                 E_proj = get_EF(input_size, dim_k)
                 F_proj = get_EF(input_size, dim_k)
@@ -152,11 +180,14 @@ class MHAttention(nn.Module):
         self.activation = get_act(activation)
 
     def forward(self, tensor, **kwargs):
+        batch_size, input_len, channels = tensor.shape
+        assert not (self.decoder_mode and "embeddings" not in kwargs), "Embeddings must be supplied if decoding"
+        assert not ("embeddings" in kwargs and (kwargs["embeddings"].shape[0], kwargs["embeddings"].shape[1], kwargs["embeddings"].shape[2]) != (batch_size, input_len, channels)), "Embeddings size must be the same as the input tensor"
         head_outputs = []
         for index, head in enumerate(self.heads):
             Q = self.to_q[index](tensor)
-            K = self.to_k[index](tensor)
-            V = self.to_v[index](tensor)
+            K = self.to_k[index](tensor) if not self.decoder_mode else self.to_k[index](kwargs["embeddings"])
+            V = self.to_v[index](tensor) if not self.decoder_mode else self.to_v[index](kwargs["embeddings"])
             if self.checkpoint_level == "C2":
                 head_outputs.append(checkpoint(head,Q,K,V))
             else:
@@ -176,34 +207,54 @@ class Linformer(nn.Module):
     My attempt at reproducing the Linformer Paper
     https://arxiv.org/pdf/2006.04768.pdf
     """
-    def __init__(self, input_size, channels, dim_k, dim_ff=256, dim_d=None, dropout_ff=0.15, nhead=4, depth=1, dropout=0.1, activation="gelu", use_pos_emb=True, checkpoint_level="C0", parameter_sharing="layerwise", k_reduce_by_layer=0, full_attention=False, include_ff=True, w_o_intermediate_dim=None):
+    def __init__(self, input_size, channels, dim_k, dim_ff=256, dim_d=None, dropout_ff=0.15, nhead=4, depth=1, dropout=0.1, activation="gelu", checkpoint_level="C0", parameter_sharing="layerwise", k_reduce_by_layer=0, full_attention=False, include_ff=True, w_o_intermediate_dim=None, decoder_mode=False):
         super(Linformer, self).__init__()
         assert activation == "gelu" or activation == "relu", "Only gelu and relu activations supported for now"
         assert checkpoint_level == "C0" or checkpoint_level == "C1" or checkpoint_level == "C2", "Checkpoint level has to be either C0, C1, or C2."
         assert parameter_sharing == "none" or parameter_sharing == "headwise" or parameter_sharing == "kv" or parameter_sharing == "layerwise", "The `parameter_sharing` flag has to be either 'none', 'headwise', 'kv', or 'layerwise'."
         assert channels % nhead == 0 if dim_d is None else True, "If `dim_d` is not set to a custom value, `channels` must be divisible by `nhead`!"
 
-        self.layers = nn.ModuleList()
+        layers = nn.ModuleList()
+        self.decoder_mode = decoder_mode
         self.input_size = input_size
         self.channels = channels
         self.checkpoint_level = checkpoint_level
-        self.pos_emb = PositionalEmbedding(channels) if use_pos_emb else None
         self.depth = depth
         self.nhead = nhead
 
         head_dim = channels // nhead if dim_d is None else dim_d
 
-        self.E = get_EF(input_size, dim_k)
-        self.F = self.E
+        E_proj = get_EF(input_size, dim_k)
 
-        get_attn = lambda curr_dim_k: MHAttention(input_size, head_dim, channels, curr_dim_k, nhead, dropout, activation, checkpoint_level, parameter_sharing, self.E, self.F, full_attention, w_o_intermediate_dim)
+        get_attn = lambda curr_dim_k: MHAttention(input_size, head_dim, channels, curr_dim_k, nhead, dropout, activation, checkpoint_level, parameter_sharing, E_proj, E_proj, full_attention, w_o_intermediate_dim, decoder_mode=False)
+        get_attn_context = lambda curr_dim_k: MHAttention(input_size, head_dim, channels, curr_dim_k, nhead, dropout, activation, checkpoint_level, parameter_sharing, E_proj, E_proj, full_attention, w_o_intermediate_dim, decoder_mode=True)
         get_ff = lambda: FeedForward(channels, dim_ff, dropout_ff)
-        get_norm = lambda: nn.LayerNorm(channels)
 
         for index in range(depth):
-            self.layers.append(nn.ModuleList([get_attn(max(1, dim_k - index*k_reduce_by_layer)), get_norm()]))
+            attn_layer = get_attn(max(1, dim_k - index*k_reduce_by_layer))
+            ff_layer = get_ff()
+
+            attn_layer, ff_layer = map(lambda fn: Residual(PreNorm(channels, fn)), (attn_layer, ff_layer))
+
             if include_ff:
-                self.layers.append(nn.ModuleList([get_ff(), get_norm()]))
+                layers.extend([attn_layer, ff_layer])
+            else:
+                layers.extend([attn_layer])
+
+            if not self.decoder_mode:
+                continue
+
+            attn_context = get_attn_context(max(1, dim_k - index*k_reduce_by_layer))
+            ff_context = get_ff()
+
+            attn_context, ff_context = map(lambda fn: Residual(PreNorm(channels, fn)), (attn_context, ff_context))
+
+            if include_ff:
+                layers.extend([attn_context, ff_context])
+            else:
+                layers.extend([attn_context])
+
+        self.seq = layers
 
     def forward(self, tensor, **kwargs):
         """
@@ -212,30 +263,21 @@ class Linformer(nn.Module):
         bt, n, c = tensor.shape
         assert n == self.input_size, "This tensor is of the wrong size. Dimension 1 has to match the `input_size` flag"
         assert c == self.channels, "This tensor is of the wrong size. Dimension 2 has to match the `channels` flag"
-        assert self.checkpoint_level == "C0" if kwargs else True, "Cannot run checkpointing when visualizing. Please set the checkpoint level to `C0`"
+        assert self.checkpoint_level == "C0" if kwargs else True, "Cannot run checkpointing when using kwargs. Please set the checkpoint level to `C0`"
+        assert "embeddings" not in kwargs or self.decoder_mode, "If decoding, needs to be initialized with `decoder_mode=True`"
 
-        if self.pos_emb is not None:
-            tensor += self.pos_emb(tensor).type(tensor.type())
-
-        for layer in self.layers:
-            module = layer[0]
-            norm = layer[1]
-
-            before_module_tensor = tensor.clone().detach()
-            if self.checkpoint_level == "C1" or self.checkpoint_level == "C2":
-                tensor = checkpoint(module, tensor)
+        for layer in self.seq:
+            if self.checkpoint_level != "C0":
+                tensor = checkpoint(layer, tensor)
             else:
-                tensor = module(tensor, **kwargs)
-            tensor += before_module_tensor
-            tensor = norm(tensor)
-
+                tensor = layer(tensor, **kwargs)
         return tensor
 
 class LinformerLM(nn.Module):
     """
     A wrapper function to accept LM tasks, inspired by https://github.com/lucidrains/sinkhorn-transformer
     """
-    def __init__(self, num_tokens, input_size, channels, dim_k=64, dim_ff=1024, dim_d=None, dropout_ff=0.1, nhead=4, depth=2, dropout=0.05, activation="gelu", checkpoint_level="C0", parameter_sharing="layerwise", k_reduce_by_layer=0, full_attention=False, include_ff=True, w_o_intermediate_dim=None, emb_dim=None):
+    def __init__(self, num_tokens, input_size, channels, dim_k=64, dim_ff=1024, dim_d=None, dropout_ff=0.1, nhead=4, depth=2, dropout=0.05, activation="gelu", checkpoint_level="C0", parameter_sharing="layerwise", k_reduce_by_layer=0, full_attention=False, include_ff=True, w_o_intermediate_dim=None, emb_dim=None, return_emb=False, decoder_mode=False):
         super(LinformerLM, self).__init__()
         emb_dim = channels if emb_dim is None else emb_dim
 
@@ -243,12 +285,12 @@ class LinformerLM(nn.Module):
 
         self.to_token_emb = nn.Embedding(num_tokens, emb_dim)
         self.pos_emb = PositionalEmbedding(emb_dim)
-        self.linformer = Linformer(input_size, channels, dim_k=dim_k, dim_ff=dim_ff, dim_d=dim_d, dropout_ff=dropout_ff, nhead=nhead, depth=depth, dropout=dropout, activation=activation, use_pos_emb=False, checkpoint_level=checkpoint_level, parameter_sharing=parameter_sharing, k_reduce_by_layer=k_reduce_by_layer, full_attention=full_attention, include_ff=include_ff, w_o_intermediate_dim=w_o_intermediate_dim)
+        self.linformer = Linformer(input_size, channels, dim_k=dim_k, dim_ff=dim_ff, dim_d=dim_d, dropout_ff=dropout_ff, nhead=nhead, depth=depth, dropout=dropout, activation=activation, checkpoint_level=checkpoint_level, parameter_sharing=parameter_sharing, k_reduce_by_layer=k_reduce_by_layer, full_attention=full_attention, include_ff=include_ff, w_o_intermediate_dim=w_o_intermediate_dim, decoder_mode=decoder_mode)
 
         if emb_dim != channels:
             self.linformer = ProjectInOut(self.linformer, emb_dim, channels)
 
-        self.to_logits = nn.Linear(emb_dim, num_tokens)
+        self.to_logits = identity if return_emb else nn.Linear(emb_dim, num_tokens)
 
     def forward(self, tensor, **kwargs):
         """
